/* lslp.c - a C version of Barrodale and Erickson
		least-squares linear prediction
		for maximum entropy spectrum estimation */

#include <stdio.h>
#include <math.h>
#include "nrutil.h"
#include "complex.h"
#include "getargs.h"


#define		YES		-1
#define		NO		0


/* function prototypes */
void LinearRegression( float *y, int n, float *slope, float *intcpt );
void SubtractBkgnd( float *y, int n );



int findfreq = YES;		/* use least-squares to find fundamental frequency */
float givenfreq = 0.06;		/* default frequency to use if !findfreq */
int m = 5;			/* max order of analysis */
int spectrum = NO;		/* print the power spectrum to stdout */
int prphase = NO;		/* print phase to stdout */
int prmag = NO;			/* print magnitude to stdout */
int prdata = NO;		/* print modified data to stdout */
int prfit = NO;			/* print fit to data to stdout */
int prfreq = NO;		/* print frequency */

ARG arglist[] = {
	{ "l",		BOOLEAN,	&findfreq,		"use linear prediction for frequency"	},
	{ "f",		FLOAT,		(int *)&givenfreq,	"default frequency (if no liner pred)"	},
	{ "m",		INTEGER,	&m,			"order of analysis"			},
	{ "spect",	BOOLEAN,	&spectrum,		"print power spectrum"			},	
	{ "phase",	BOOLEAN,	&prphase,		"print phase"				},	
	{ "mag",	BOOLEAN,	&prmag,			"print magnitude"			},	
	{ "data",	BOOLEAN,	&prdata,		"print modified data"			},	
	{ "fit",	BOOLEAN,	&prfit,			"print fit to data"			},	
	{ "freq",	BOOLEAN,	&prfreq,		"print frequency"			},	
};

#define		NARGS		(sizeof( arglist ) / sizeof( ARG ))




#define		LSLP_FORWARD		0
#define		LSLP_BACKWARD		1
#define		LSLP_FORBACK		2

#define		NMAX		2000
#define		MMAX		500



void main( argc, argv )
int argc;
char *argv[];
{
int i, j;
float ave;
float *y, *a;
fcomplex ca[MMAX+1], poles[MMAX+1];
float yy;
float PredictionResidual( );
int n;
FILE *fp;
float fundamental, frequency(), mag, phase;


	/* process command line arguments */
	argc = getargs( argc, argv, arglist, NARGS );
	if ( argc != 2 ) {
		fprintf( stderr, "Usage:  %s [switches] DataFile\n", argv[0] );
		pr_usage( arglist, NARGS );
		exit( 0 );
	}
	if ( m < 1 || m > MMAX ) {
		fprintf( stderr, "Error:  illegal value for -m\n" );
		exit( 0 );
	}
	if ( (fp = fopen( argv[1], "r")) == NULL ) {
		fprintf( stderr, "Error:  cannot open %s for read\n", argv[2] );
		exit( 0 );
	}




	/* read data and subtract linear background */
	y = vector( 1, NMAX );
	n = 0;
	while ( fscanf( fp, "%f", &yy ) == 1 ) {
		++n;
		y[n] = yy;
	}
	fprintf( stderr, "Read %d values from <%s>\n", n, argv[1] );
	fclose( fp );
	SubtractBkgnd( y, n );


	/* set frequency */
	if ( findfreq ) {
		a = vector( 1, m );
		LeastSquaresLinearPrediction( y, n, a, m, LSLP_FORBACK );
		if ( spectrum )
			PrintPowerSpectrum( stdout, y, n, a, m, LSLP_FORBACK, 100 );

		/* compute the fundamental frequency */
		fundamental = frequency( a, m );
	} else
		fundamental = givenfreq;
	fprintf( stderr, "fundamental frequency = %g(%s)\n", fundamental,
		findfreq ? "least-squares linear prediction" : "default" );
	if ( prfreq )
		fprintf( stdout, "%g\n", fundamental );


	/* find the magnitude and phase of the fundamental component */
	MagPhase( y, n, fundamental, &mag, &phase );
	fprintf( stderr, "Magnitude = %g  Phase = %g\n", mag, phase / (2.0 * M_PI) );
	if ( prphase )
		fprintf( stdout, "%g\n", phase / (2.0 * M_PI) );
	if ( prmag )
		fprintf( stdout, "%g\n", mag );

	if ( prdata )
		for( i = 1; i <= n; i++ )
			fprintf( stdout, "%g\n", y[i] );

	if ( prfit )
		for( i = 1; i <= n; i++ )
			fprintf( stdout, "%g\n",
				1.41 * mag * cos( -2.0 * M_PI * fundamental * (i-1) - phase ) );
}



LeastSquaresLinearPrediction( x, n, a, m, mode )
float *x;	/* time series data indexed from 1 to n */
int n;		/* number of samples in time series */
float *a;	/* prediction coefficients computed indexed from 1 to m */
int m;		/* number of coefficients */
int mode;	/* mode Forward - 0   Backward - 1   Forward and Backward - 2 */
{
float **r;	/* auto correlation matrix */
float *s;	/* rhs vector of linear system */
int *indx;	/* required by ludcmp */
float d;	/* required by ludcmp */
int i, j, t;
float sum1, sum2;

	/* the components of the linear system to solve */
	r = matrix( 1, m, 1, m );
	s = a;	/* lubksb returns solution over top of r.h.s. vector */

	/* vector required by ludcmp */
	indx = ivector( 1, m );

	/* set up the symetric positive definite matrix */
	for ( i = 1; i <= m; i++ )
		for( j = i; j <= m; j++ ) {
			sum1 = sum2 = 0.0;
			for( t = 1; t <= (n-m); t++ ) {
				sum1 += x[m+t-j] * x[m+t-i];
				sum2 += x[t+j] * x[t+i];
			}
			if ( mode == LSLP_FORWARD )
				r[i][j] = sum1;
			else if (mode == LSLP_BACKWARD )
				r[i][j] = sum2;
			else if ( mode == LSLP_FORBACK )
				r[i][j] = sum1 + sum2;
			else {
				fprintf( stderr, "LeastSquaresLinearPrediction: error - illegal mode\n" );
				return;
			}

			if ( i != j ) r[j][i] = r[i][j];
		}

	/* set up the r.h.s. vector */
	for( i = 1; i <= m; i++ ) {
		sum1 = sum2 = 0.0;
		for( t = 1; t <= (n-m); t++ ) {
			sum1 += x[m+t-i] * x[m+t];
			sum2 += x[t+i] * x[t];
		}
		if ( mode == LSLP_FORWARD )
				s[i] = sum1;
			else if (mode == LSLP_BACKWARD )
				s[i] = sum2;
			else if ( mode == LSLP_FORBACK )
				s[i] = sum1 + sum2;
			else {
				fprintf( stderr, "LeastSquaresLinearPrediction: error - illegal mode\n" );
				return;
			}
	}

	ludcmp( r, m, indx, &d );
	lubksb( r, m, indx, s );

	free_ivector( indx, 1, m );
	free_matrix( r, 1, m, 1, m );
}


float PredictionResidual( x, n, a, m, mode )
float *x;
int n;
float *a;
int m;
int mode;
{
int i, j, k;
float sum, e;

	sum = 0.0;

	if ( mode == LSLP_FORWARD || mode == LSLP_FORBACK ) {
		for( i = m+1; i <= n; i++ ) {
			e = 0.0;
			for( j = i - 1, k = 1; j >= i - m; j--, k++ )
				e += a[k] * x[j];
			e -= x[i];
			sum += e * e;
		}
	}

	if ( mode == LSLP_BACKWARD || mode == LSLP_FORBACK ) {
		for( i = 1; i <= n-m; i++ ) {
			e = 0.0;
			for( j = i + 1, k = 1; j <= i + m; j++, k++ )
				e += a[k] * x[j];
			e -= x[i];
			sum += e * e;
		}
	}


	return sum;
}

PrintPowerSpectrum( fp, x, n, a, m, mode, npts )
FILE *fp;
float *x;
int n;
float *a;
int m;
int mode;
int npts;
{
int i, j;
float w, interval;
float sumre, sumim, power;
float ssq, k;


	interval = 0.5 / (float) npts;
	ssq = PredictionResidual( x, n, a, m, mode );
	k = ssq / (2.0 * (n - m));
	
	for( i = 0; i <= npts; i++ ) {
		w = 2.0 * M_PI * i * interval;
		sumre = 1.0;
		sumim = 0.0;
		for( j = 1; j <= m; j++ ) {
			sumre -= a[j] * cos( j * w );
			sumim -= a[j] * sin( j * w );
		}
		power = k / (sumre * sumre + sumim * sumim);

		fprintf( fp, "%g %g\n", i * interval, power );
	}

}


#define		NPTS		1000

float frequency( a, m )
float *a;
int m;
{
int i, j;
float w, interval;
float sumre, sumim, power;
float min, wmin;

	interval = 0.5 / (float) NPTS;

	wmin = w = 0.0;
	sumre = 1.0;
	sumim = 0.0;
	for( j = 1; j <= m; j++ ) {
		sumre -= a[j] * cos( j * w );
		sumim -= a[j] * sin( j * w );
	}
	min = sumre * sumre + sumim * sumim;


	
	for( i = 1; i <= NPTS; i++ ) {
		w = 2.0 * M_PI * i * interval;
		sumre = 1.0;
		sumim = 0.0;
		for( j = 1; j <= m; j++ ) {
			sumre -= a[j] * cos( j * w );
			sumim -= a[j] * sin( j * w );
		}
		power = sumre * sumre + sumim * sumim;

		if ( power < min ) {
			min = power;
			wmin = i * interval;
		}
	}

	return wmin;
}

MagPhase( x, n, freq, mag, phase )
float *x;
int n;
float freq, *mag, *phase;
{
int i;
float cs, sn, re, im;

	re = im = 0.0;
	for( i = 1; i <= n; i++ ) {
		cs = cos( -2.0 * M_PI * freq * (i-1) );
		sn = sin( -2.0 * M_PI * freq * (i-1) );
		re += x[i] * cs;
		im += x[i] * sn;
	}
	re /= (float)n;
	im /= (float)n;

	*mag = sqrt( re * re + im * im );
	if ( re == 0.0 ) {
		if ( im >= 0.0 ) *phase = M_PI / 2.0;
			else *phase = -M_PI / 2.0;
	} else
		*phase = atan2( im, re );

}


void LinearRegression( float *y, int n, float *slope, float *intcpt )
{
float sumx, sumy, sumxx, sumyy, sumxy;
int i;

	sumx = sumy = sumxx = sumyy = sumxy = 0.0;
	for( i = 1; i <= n; i++ ) {
		sumx += i;
		sumy += y[i];
		sumxx += i * i;
		sumyy += y[i] * y[i];
		sumxy += i * y[i];
	}

	*slope = (n * sumxy - sumx * sumy) / (n * sumxx - sumx * sumx);
	*intcpt = (sumy / n) - *slope * (sumx / n);
}

void SubtractBkgnd( float *y, int n )
{
float slope, intcpt;
int i;

	LinearRegression( y, n, &slope, &intcpt );

	for( i = 1; i <= n; i++ )
		y[i] -= (slope * i + intcpt);
}
